using System;
using System.Collections.Generic;
using System.Linq;
using static Keysight.Hvi.ProgrammingExamples.ProgExample5.ProgramExample5;

namespace Keysight.Hvi.ProgrammingExamples.ProgExample5
{
    /// <summary>
    /// Main Program
    /// </summary>
    class Program
    {
        static void Main(string[] args)
        {
            IHvi hvi = null;
            List<SD1AwgModule> moduleList = null;

            try
            {
                #region Open PXI modules, setup AWGs and queue waveforms

                // Load configuration
                var appConfig = new ApplicationConfig();

                // Checks flags defining if the code executes on HW or Simulation mode
                if (ApplicationConfig.HardwareSimulated)
                {
                    Console.WriteLine("Code running in Simulation Mode");
                }

                // Open instruments
                moduleList = SD1AwgModule.OpenModules();

                // Queue AWG Waveforms
                moduleList.ForEach(module => module.ConfigureAWG(appConfig));

                #endregion

                #region System definition

                var mySystem =
                    DefineSystem("mySystem") // Define your system, HW platform, add HVI resources
                    .AddHwPlatform(appConfig)  // Add chassis, interconnections, PXI trigger resources, synchronization, HVI clocks
                    .AddHviEngines(moduleList) // Define all the HVI engines to be included in the HVI
                    .AddHviActions(moduleList, appConfig) // Define list of actions to be executed
                    .AddHviTriggers(moduleList) // Defines the trigger resources
                    .OutputSystemDef(); // Output to console the details of the system definition

                #endregion

                #region Program HVI sequences

                Console.WriteLine("-----------------------");
                Console.WriteLine("Program HVI Sequences");
                Console.WriteLine("-----------------------");

                // Create and program sequencer object
                Console.WriteLine("Initializing the defined system...");
                var sequencer =
                    DefineSequencer("MySequencer", mySystem)
                    .ProgramMixedSignalSequence(moduleList, appConfig);// Program HVI sequences using the sequencer object
                Console.WriteLine("Programming the HVI sequences...");

                // Export the programmed HVI sequences to text
                ExportHviSequences(sequencer);

                #endregion

                #region Compile, Load to HW, Run HVI Sequence

                Console.WriteLine("-------------------");
                Console.WriteLine("Execute HVI");
                Console.WriteLine("-------------------");

                // Compile HVI sequences
                Console.WriteLine("Compiling HVI...");
                hvi = sequencer.Compile();
                Console.WriteLine(hvi.CompileStatus.ToString()); // Prints warning end error messages
                Console.WriteLine("HVI Compiled");
                Console.WriteLine("This HVI needs to reserve {0} PXI trigger resources to execute", hvi.CompileStatus.SyncResources.Count());

                // Load HVI to HW: load sequences, configure actions/triggers/events, lock resources, etc.
                hvi.LoadToHw();
                Console.WriteLine("HVI Loaded to HW");

                // Execute HVI in blocking mode: SW waits until HVI sequences ends their execution
                // Eventually enter a timeout for the HVI execution to be stopped: timeout = timedelta(seconds=0), hvi.run(timeout)
                hvi.Run(IHvi.NoTimeout);
                Console.WriteLine("HVI Running...");

                // User-controlled HVI execution
                if (ApplicationConfig.HardwareSimulated)
                {
                    Console.WriteLine("Simulation completed successfully");
                }
                else
                {
                    InteractWithHvi(hvi);
                }

                #endregion
            }
            finally
            {
                #region Release HW and Close instrument objects

                // Release HW resources once HVI execution is completed
                hvi?.ReleaseHw();
                Console.WriteLine("Releasing HW...");

                // Close all modules at the end of the execution
                SD1AwgModule.CloseModules(moduleList);
                Console.WriteLine("Modules closed\n");
                Console.WriteLine("Press a key to exit...");
                var inputKey = Console.ReadLine();
                #endregion
            }
        }

        /// <summary>
        /// Method that allows the user to control the HVI execution and interact with it 
        /// </summary>
        /// <param name="hvi">Hard Virtual Instrument generated by compiling the sequencer object.</param>
        static void InteractWithHvi(IHvi hvi)
        {
            // If running with hardware, iteratively re-run HVI until the user decides to exit
            while (true)
            {
                Console.WriteLine("Press enter to run HVI again, q to exit...");
                string inputKey = Console.ReadLine();
                if (inputKey == "q")
                    break;
                hvi.Run(IHvi.NoTimeout);
            }
        }

        /// <summary>
        /// Utility method to handle null objects.
        /// </summary>
        /// <param name="obj">Object to be handled.</param>
        /// <param name="objName">Object name.</param>
        public static void AssertNotNull(object obj, string objName)
        {
            if (obj == null)
            {
                throw new ArgumentNullException(objName);
            }
        }
    }
}
