using System;
using System.Collections.Generic;
using static Keysight.Hvi.ProgrammingExamples.ProgExample5.Program;

namespace Keysight.Hvi.ProgrammingExamples.ProgExample5
{
    /// <summary>
    /// Contains the system definition methods for Prog. Example 5
    /// Defines the system resources to be used by HVI
    /// </summary>
    public static partial class ProgramExample5
    {
        /// <summary>
        /// Initialize the system
        /// </summary>
        /// <param name="mySystem"></param>
        /// <returns>Outputs the details of each instrument and engine included in the system definition</returns>
        public static SystemDefinition DefineSystem(string name)
        {
            return new SystemDefinition(name);
        }

        /// <summary>
        /// Define HW platform: chassis, interconnections, PXI trigger resources, synchronization, HVI clocks
        /// </summary>
        /// <param name="mySystem">SystemDefinition object defining the system where HVI runs.</param>
        /// <param name="appConfig">User defined configuration.</param>
        /// <returns>SystemDefinition object defining the system where HVI runs.</returns>
        public static SystemDefinition AddHwPlatform(this SystemDefinition mySystem, ApplicationConfig appConfig)
        {
            // Check input parameters
            AssertNotNull(mySystem, nameof(mySystem));
            AssertNotNull(appConfig, nameof(appConfig));

            //Add chassis resources
            // For multi-chassis setup details see programming example documentation
            foreach (int chassisNumber in appConfig.ChassisList)
            {
                if (ApplicationConfig.HardwareSimulated)
                {
                    mySystem.Chassis.AddWithOptions(chassisNumber, "Simulate=True,DriverSetup=Model=GenericPxieChassis");
                }
                else
                {
                    mySystem.Chassis.Add(chassisNumber);
                }
            }

            // Define System Sync Modules (SSMs)
            if (appConfig.SystemSyncModulesDescriptors != null)
            {
                var interconnects = mySystem.Interconnects;
                List<Stm.ISyncModule> ssmList = new List<Stm.ISyncModule>();
                foreach (var descriptor in appConfig.SystemSyncModulesDescriptors)
                {
                    if (ApplicationConfig.HardwareSimulated == true)
                        ssmList.Add(interconnects.AddSyncModule(descriptor.ResourceId, ApplicationConfig.SsmSimulationOptions));
                    else
                        ssmList.Add(interconnects.AddSyncModule(descriptor.ResourceId, descriptor.Options));
                }
                // Define connections between SSMs
                if (appConfig.SsmConnections != null)
                {
                    foreach (var connection in appConfig.SsmConnections)
                    {
                        int connectorNumber = connection.Ssm1DownstreamConnectorNumber;
                        Stm.ISyncModule ssm1 = null;
                        Stm.ISyncModule ssm2 = null;
                        foreach (var ssm in ssmList)
                        {
                            if (ssm.Chassis == connection.Ssm1Chassis)
                                ssm1 = ssm;
                            if (ssm.Chassis == connection.Ssm2Chassis)
                                ssm2 = ssm;
                        }
                        // Implement each user-defined connection
                        try
                        {
                            // Set connection. SSMs have always one upstream port
                            ssm1.Connectivity.SystemSyncDown[connectorNumber].SetConnection(ssm2.Connectivity.SystemSyncUp[0]);
                        }
                        catch
                        {
                            Console.WriteLine("Exception! Please check the valued defined for SyncModule resource ids, chassis numbers and connections");
                        }
                    }
                }
            }

            // Assign the defined PXI trigger resources
            mySystem.SyncResources = appConfig.PxiSyncTriggerResources;

            // Assign clock frequencies that are outside the set of the clock frequencies of each HVI engine
            // Use the code line below if you want the application to be in sync with the 10 MHz clock
            mySystem.NonHviCoreClocks = new double[] { 10e6 };

            return mySystem;
        }

        /// <summary>
        /// Define all the HVI engines to be included in the HVI
        /// </summary>
        /// <param name="mySystem">SystemDefinition object defining the system where HVI runs.</param>
        /// <param name="moduleList">List of objects representing modular instruments.</param>
        /// <returns>SystemDefinition object defining the system where HVI runs.</returns>
        public static SystemDefinition AddHviEngines(this SystemDefinition mySystem, List<SD1AwgModule> moduleList)
        {
            // Check input parameters
            AssertNotNull(mySystem, nameof(mySystem));
            AssertNotNull(moduleList, nameof(moduleList));

            // For each instrument to be used in the HVI application add its HVI Engine to the HVI Engine Collection 
            foreach (var module in moduleList)
            {
                mySystem.Engines.Add(module.Instrument.Hvi.Engines.MainEngine, module.HVIEngineName);
            }

            return mySystem;
        }

        /// <summary>
        /// Defines AWG trigger actions for each module, to be executed by the "action execute" instruction in the HVI sequence
        /// </summary>
        /// <param name="mySystem">SystemDefinition object defining the system where HVI runs.</param>
        /// <param name="moduleList">List of objects representing modular instruments.</param>
        /// <param name="appConfig">User defined configuration.</param>
        /// <returns>SystemDefinition object defining the system where HVI runs.</returns>
        public static SystemDefinition AddHviActions(this SystemDefinition mySystem, List<SD1AwgModule> moduleList, ApplicationConfig appConfig)
        {
            // Check input parameters
            AssertNotNull(mySystem, nameof(mySystem));
            AssertNotNull(moduleList, nameof(moduleList));
            AssertNotNull(appConfig, nameof(appConfig));

            // For each AWG, define the list of HVI Actions to be executed and add such list to its own HVI Action Collection
            foreach (var module in moduleList)
            {
                for (var chIndex = 1; chIndex <= module.NumChannels; chIndex++)
                {
                    // Actions need to be added to the engine's action list so that they can be executed
                    string actionName = string.Format("{0}{1}", ApplicationConfig.AwgTriggerActionName, chIndex);  // arbitrary user-defined name
                    int actionId;
                    switch (chIndex)
                    {
                        case 2:
                            actionId = module.Instrument.Hvi.Actions.Awg2Trigger;
                            break;
                        case 3:
                            actionId = module.Instrument.Hvi.Actions.Awg3Trigger;
                            break;
                        case 4:
                            actionId = module.Instrument.Hvi.Actions.Awg4Trigger;
                            break;
                        default:
                            actionId = module.Instrument.Hvi.Actions.Awg1Trigger;
                            break;
                    }
                    mySystem.Engines[module.HVIEngineName].Actions.Add(actionId, actionName);
                }
            }

            return mySystem;
        }

        /// <summary>
        /// Defines and configure the FP trigger output of each AWG
        /// </summary>
        /// <param name="mySystem">SystemDefinition object defining the system where HVI runs.</param>
        /// <param name="moduleList">List of objects representing modular instruments.</param>
        /// <param name="appConfig">User defined configuration.</param>
        /// <returns>SystemDefinition object defining the system where HVI runs.</returns>
        public static SystemDefinition AddHviTriggers(this SystemDefinition mySystem, List<SD1AwgModule> moduleList)
        {
            // Check input parameters
            AssertNotNull(mySystem, nameof(mySystem));
            AssertNotNull(moduleList, nameof(moduleList));

            // Add to the HVI Trigger Collection of each HVI Engine the FP Trigger object of that same instrument
            foreach (var module in moduleList)
            {
                int fpTriggerId = module.Instrument.Hvi.Triggers.FrontPanel1;
                ITriggerDefinition fpTrigger = mySystem.Engines[module.HVIEngineName].Triggers.Add(fpTriggerId, ApplicationConfig.FpTriggerName);
                
                // Configure FP trigger in each Hvi Engine
                fpTrigger.Config.Direction = Direction.Output;
                fpTrigger.Config.Polarity = Polarity.ActiveHigh;
                fpTrigger.Config.SyncMode = SyncMode.Immediate;
                fpTrigger.Config.HwRoutingDelay = 0;
                
                // FP TriggerMode is set to Level, which does not defines a pulse length
                // FP trigger pulse length is defined by the HVI Statements that control FP Trigger ON/OFF
                fpTrigger.Config.TriggerMode = TriggerMode.Level;
            }

            return mySystem;
        }

        /// <summary>
        /// Outputs the details of each instrument and engine included in the system definition.
        /// </summary>
        /// <param name="mySystem">SystemDefinition object defining the system where HVI runs.</param>
        /// <returns>SystemDefinition object defining the system where HVI runs.</returns>
        public static SystemDefinition OutputSystemDef(this SystemDefinition mySystem)
        {
            Console.WriteLine("-------------------");
            Console.WriteLine("System Definition");
            Console.WriteLine("-------------------");
            // Output KS2201A HVI Core Version
            IVersion hviCoreVersion = SystemDefinition.HviCoreVersion;
            Console.WriteLine("The PathWave Test Sync Executive API installed on your system has an HVI core version " + hviCoreVersion.ToString() + "\n");
            // Output info about the HW instruments
            Console.WriteLine("HW Instruments:");
            foreach (var descriptor in ApplicationConfig.ModuleDescriptors)
            {
                var engine = mySystem.Engines[descriptor.EngineName];
                Console.WriteLine("- Model: {0}, HVI Engine Name: {1}, HVI core version: " +
                engine.SoftwareVersion.ToString() + ", HVI Engine FPGA IP version: " + engine.FirmwareVersion.ToString(), descriptor.ModelNumber, descriptor.ApplicationRole);
            }
            // Output info about the Sync Modules
            if (mySystem.Interconnects != null)
            {
                Console.WriteLine("System Sync Modules:");
                for (int index = 0; index < mySystem.Interconnects.Count; index++)
                {
                    var syncModule = mySystem.Interconnects[index];
                    Console.WriteLine("- System Sync Module in chassis: {0}, slot: {1}", syncModule.Chassis, syncModule.Slot);
                }
            }
            return mySystem;
        }
    }
}
