using System;
using System.Collections.Generic;
using System.Linq;
using KeysightSD1;
using static Keysight.Hvi.ProgrammingExamples.ProgExample5.Program;

namespace Keysight.Hvi.ProgrammingExamples.ProgExample5
{
    /// <summary>
    /// Class to use SD1/M3xxxA Modular Instruments
    /// </summary>
    public class SD1AwgModule
    {
        public SD_AOU Instrument { set; get; }
        public short NumChannels { set; get; } = 2;
        public string HVIEngineName { set; get; } = "";

        /// <summary>
        /// Opens and creates all the necessary instrument objects.
        /// Returns a dictionary of module objects whose keys are the HVI engine names.
        /// Please check SD1 3.x User Guide for options to open the instrument objectso
        /// </summary>
        /// <returns>List of objects representing modular instruments.</returns>
        public static List<SD1AwgModule> OpenModules()
        {

            List<SD1AwgModule> moduleList = new List<SD1AwgModule>();
            int moduleCount = 0, awgCount = 0, dioCount = 0;

            foreach (var descriptor in ApplicationConfig.ModuleDescriptors)
            {
                var instrObject = new SD_AOU();
                string instrOptions = descriptor.Options;
                if (ApplicationConfig.HardwareSimulated)
                {
                    instrOptions = descriptor.Options + ",simulate=true";
                }
                var id = instrObject.open(descriptor.ModelNumber, descriptor.ChassisNumber, descriptor.SlotNumber, instrOptions);
                if (id < 0)
                {
                    var errorMessage = string.Format("Error opening module in chassis: {0}, slot: {1}, opened with ID: {2}",
                        descriptor.ChassisNumber, descriptor.SlotNumber, id);
                    throw new Exception(errorMessage);
                }

                //Get number of channels
                short numChannels;
                switch (instrObject.getOptions("channels"))
                {
                    case "CH2":
                        numChannels = 2;
                        break;
                    case "CH4":
                        numChannels = 4;
                        break;
                    default:
                        numChannels = 0;
                        break;
                }

                // Set HVI Engine name
                string engineName = "";
                if (descriptor.ApplicationRole == ApplicationConfig.DioEngineName)
                {
                    engineName = String.Format("{0}{1}", ApplicationConfig.DioEngineName, dioCount);
                    dioCount++;
                }
                else if (descriptor.ApplicationRole == ApplicationConfig.AwgEngineName)
                {
                    engineName = String.Format("{0}{1}", ApplicationConfig.AwgEngineName, awgCount);
                    awgCount++;
                }

                //Create Module object               
                moduleList.Add(new SD1AwgModule
                {
                    Instrument = instrObject,
                    NumChannels = numChannels,
                    HVIEngineName = engineName,
                });
                descriptor.EngineName = engineName;
                moduleCount++;
            }

            // Choose HVI engine to act as Leader Engine
            ApplicationConfig.LeaderEngineName = moduleList.First().HVIEngineName;

            return moduleList;
        }

        /// <summary>
        /// Closes the instrument for each module in the list0
        /// </summary>
        /// <param name="modulesList">List of objects representing modular instruments.</param>
        public static void CloseModules(List<SD1AwgModule> modulesList)
        {
            AssertNotNull(modulesList, nameof(modulesList));

            modulesList.ForEach(module => module?.Instrument?.close());
        }

        /// <summary>
        /// ConfigureAWG queues a Gaussian waveform to all the channels 
        /// of the AWG object awgModule that is passed to the function  
        /// </summary>
        /// <param name="appConfig">User defined configuration.</param>
        public void ConfigureAWG(ApplicationConfig appConfig)
        {
            // Check application configuration
            AssertNotNull(appConfig, nameof(appConfig));

            // Load waveform to AWG memory
            Instrument.waveformFlush(); //memory flush
            var wave = new SD_Wave(appConfig.WaveformPath);
            int wfmNum = 0;
            Instrument.waveformLoad(wave, wfmNum);
            for (int channel = 1; channel < NumChannels; channel++)
            {
                //AWG queue flush 
                Instrument.AWGstop(channel);
                Instrument.AWGflush(channel);
                //Set AWG mode
                Instrument.channelWaveShape(channel, SD_Waveshapes.AOU_AWG);
                Instrument.channelAmplitude(channel, appConfig.Amplitude);
                //AWG configuration
                Instrument.AWGqueueConfig(channel, appConfig.QueueMode);
                Instrument.AWGqueueSyncMode(channel, appConfig.SyncMode);
                // Queue waveform to the selected channel
                Instrument.AWGqueueWaveform(channel, wfmNum, appConfig.AwgTriggerMode, appConfig.StartDelay, appConfig.AwgNumCycles, appConfig.AwgPrescaler);
                Instrument.AWGstart(channel); //AWG starts and waits for an AWG trigger
            }
        }
    }

    /// <summary>
    /// Descriptor class for an SD1 PXI module 
    /// </summary>
    public class ModuleDescriptor
    {
        public string ApplicationRole { get; set; }
        public string ModelNumber { get; set; }
        public int ChassisNumber { get; set; }
        public int SlotNumber { get; set; }
        public string Options { get; set; }
        public string EngineName { get; set; }
    }

    /// <summary>
    /// Descriptor class for System Sync Modules
    /// </summary>
    public class SystemSyncModuleDescriptor
    {
        public SystemSyncModuleDescriptor(string resourceId, string options)
        {
            ResourceId = resourceId;
            Options = options;
        }
        public string ResourceId { get; set; }
        public string Options { get; set; }
    }

    /// <summary>
    /// Describes the connection between two SyncModule objects ssm1 and ssm2:
    /// ssm1.Connectivity.SystemSyncDownstream[ssm1DownstreamConnectorNumber].SetConnection(ssm2.Connectivity.SystemsyncUpstream[0]
    /// </summary>
    public class SystemSyncModuleConnection
    {
        public int Ssm1Chassis { get; set; }
        public int Ssm1DownstreamConnectorNumber { get; set; }
        public int Ssm2Chassis { get; set; }
    }
}
